import os
import logging
logger = logging.getLogger(__name__)


class Detector:

    def __init__(self):
        from halludetector.config import init_config
        init_config(f'{os.path.dirname(os.path.realpath(__file__))}/../../config.json')

        from halludetector.config import (
            llm_handler, triplets_extractor, sentence_extractor, question_generator,
            retriever, checker, bertscorer, ngramscorer,
        )
        self.llm_handler = llm_handler
        self.triplets_extractor = triplets_extractor
        self.sentence_extractor = sentence_extractor
        self.question_generator = question_generator
        self.retriever = retriever
        self.checker = checker
        self.bertscorer = bertscorer
        self.ngramscorer = ngramscorer

    def ask_llm(self, *args, **kwargs):
        return self.llm_handler.ask_llm(*args, **kwargs)

    def extract_triplets(self, *args, **kwargs):
        return self.triplets_extractor.extract(*args, **kwargs)

    def extract_sentences(self, *args, **kwargs):
        return self.sentence_extractor.extract(*args, **kwargs)

    def generate_question(self, *args, **kwargs):
        return self.question_generator.generate(*args, **kwargs)

    def retrieve(self, *args, **kwargs):
        return self.retriever.retrieve(*args, **kwargs)

    def check(self, *args, **kwargs):
        return self.checker.check(*args, **kwargs)

    def similarity_bertscore(self, sentences, sampled_passages):
        try:
            return self.bertscorer.predict(
                sentences=sentences,  # list of sentences
                sampled_passages=sampled_passages,  # list of sampled passages
            )
        except Exception as e:
            logger.error(f'Bertscore failed due to {e}')

    def similarity_ngram(self, sentences, passage, sampled_passages):
        return self.ngramscorer.predict(
            passage=passage,
            sentences=sentences,  # list of sentences
            sampled_passages=sampled_passages,  # list of sampled passages
        )

    def score(self, question, answer=None, samples=None):
        raise NotImplementedError



import logging

from halludetector.detectors import (
    ChainPoll, SelfCheckGPTBertScore,
    SelfCheckGPTNGram, SelfCheckGPTMQAG,
    RefChecker, GEval, CustomDetector,
    SelfCheckGPTPrompt
)

scorer_mapping = {
    'Self-Check GPT Bert Score': SelfCheckGPTBertScore,
    'Self-Check GPT NGram': SelfCheckGPTNGram,
    'Self-Check GPT Prompt': SelfCheckGPTPrompt,
    'RefChecker': RefChecker,
    'G-Eval': GEval,
    'Chain Poll': ChainPoll,
    'Custom': CustomDetector
}

logger = logging.getLogger(__name__)


def calculate_score(method, question, answer=None, samples=None):
    scorer_class = scorer_mapping.get(method)
    if scorer_class:
        scorer = scorer_class()
    return scorer.score(question, answer, samples)



from typing import List, Union



def merge_ret(ret):
    """Merge results from multiple paragraphs"""
    if "Entailment" in ret:
        return "Entailment"
    if "Contradiction" in ret:
        return "Contradiction"
    return "Neutral"


def merge_multi_psg_ret(ret):
    """Merge results from multiple passages
    TODO: consider possible cases where the results are inconsistent.
    """
    if "Entailment" in ret:
        return "Entailment"
    if "Contradiction" in ret:
        return "Contradiction"
    return "Neutral"


class CheckerBase:
    def __init__(self, sentence_extractor) -> None:
        self.label_entailment = 'Entailment'
        self.label_neutral = 'Neutral'
        self.label_contradiction = 'Contradiction'
        self.labels = ["Entailment", "Neutral", "Contradiction"]
        self.sentence_extractor = sentence_extractor

    def check(
        self, 
        claim: str, 
        reference: Union[str, List], 
        response: str = None,
        question: str = None,
        max_reference_segment_length: int = 200, 
    ):
        ret = []
        if isinstance(reference, str):
            reference = [reference]
        for psg in reference:
            if max_reference_segment_length > 0:
                segments = self.split_text(psg, max_reference_segment_length)
            else:
                segments = [psg]
            psg_ret = self._check(
                claims=[claim] * len(segments),
                references=segments,
                response=response,
                question=question,
            )
            ret.append(merge_ret(psg_ret))
        return merge_multi_psg_ret(ret)

    def _check(
        self,
        claims: List,
        references: List,
        response: str,
        question: str = None
    ):
        raise NotImplementedError

    def split_text(self, text, segment_len=200):
        """Split text into segments according to sentence boundaries."""
        segments, seg = [], []
        sents = [[token.text for token in sent] for sent in self.sentence_extractor.extract(text)]
        for sent in sents:
            if len(seg) + len(sent) > segment_len:
                segments.append(" ".join(seg))
                seg = sent
                # single sentence longer than segment_len
                if len(seg) > segment_len:
                    # split into chunks of segment_len
                    seg = [
                        " ".join(seg[i:i + segment_len])
                        for i in range(0, len(seg), segment_len)
                    ]
                    segments.extend(seg)
                    seg = []
            else:
                seg.extend(sent)
        if seg:
            segments.append(" ".join(seg))
        return segments



from .gpt4_checker import GPT4Checker


class CheckerHandler:
    def __init__(self, sentence_extractor, llm_handler):
        self.checker = GPT4Checker(sentence_extractor, llm_handler)

    def check(self, *args, **kwargs):
        return self.checker.check(*args, **kwargs)



import os
import json
import math
import requests
import warnings
from typing import List, Dict, Union, Tuple

import rank_bm25
import diskcache
from bs4 import BeautifulSoup

SERPER_URL = "https://google.serper.dev/search"


class GoogleRetriever:
    def __init__(self, extractor, cache_dir: str = "./.cache"):
        self.bm25 = None
        self._load_key()
        cache_dir = os.path.join(cache_dir, "serper")
        self.cache = diskcache.Cache(cache_dir)
        self.extractor = extractor
    
    def _load_key(self):
        self.api_key = os.environ.get("SERPER_API_KEY", None)
        assert self.api_key is not None, \
            f"Require environment variable SERPER_API_KEY."

    def _query_google(self, query: str) -> dict:
        """Search Google using Serper API and retrieve abundant information"""
        if query in self.cache:
            return self.cache[query]
        else:
            payload = json.dumps({"q": query})
            headers = {
                "X-API-KEY": self.api_key,
                "Content-Type": "application/json"
            }
            response = requests.request(
                "POST", SERPER_URL, headers=headers, data=payload
            )
            response_dict = json.loads(response.text)
            self.cache[query] = response_dict
            return response_dict

    @staticmethod
    def _parse_results(results: dict) -> Tuple[List[dict], bool]:
        """Adapted from `FacTool` to utilize retrieved results as answers."""
        snippets = []
        with_answerbox = False
        if results.get("answerBox"):
            # This case indicates that Google has made a good answer to the question, and it's as desired to utilize this information.
            answer_box: dict = results.get("answerBox", {})
            if answer_box.get("answer"):
                element = {
                    "content": answer_box.get("answer"),
                    "source": answer_box.get("link"),
                }
                snippets = [element]
            elif answer_box.get("snippet"):
                element = {
                    "content": answer_box.get("snippet").replace("\n", " "),
                    "source": answer_box.get("link"),
                }
                snippets = [element]
            elif answer_box.get("snippetHighlighted"):
                element = {
                    "content": answer_box.get("snippetHighlighted"),
                    "source": answer_box.get("link"),
                }
                snippets = [element]
        if len(snippets) > 0:
            with_answerbox = True
        if results.get("knowledgeGraph"):
            kg: dict = results.get("knowledgeGraph", {})
            title = kg.get("title")
            entity_type = kg.get("type")
            if entity_type:
                element = {
                    "content": f"{title}: {entity_type}",
                    "source": kg.get("link"),
                }
                snippets.append(element)
            description = kg.get("description")
            if description:
                element = {"content": description, "source": kg.get("link")}
                snippets.append(element)
            for attribute, value in kg.get("attributes", {}).items():
                element = {"content": f"{attribute}: {value}", "source": kg.get("link")}
                snippets.append(element)
        for result in results["organic"][:3]:
            if "snippet" in result:
                element = {"content": result["snippet"], "source": result["link"]}
                snippets.append(element)
            for attribute, value in result.get("attributes", {}).items():
                element = {"content": f"{attribute}: {value}", "source": result["link"]}
                snippets.append(element)

        if len(snippets) == 0:
            warnings.warn("No usable google search results.")

        return snippets, with_answerbox

    @staticmethod
    def _get_url_text(url) -> str:
        # Read page and return text
        buf = []
        try:
            soup = BeautifulSoup(
                requests.get(url, timeout=10).text, "html.parser"
            )
            for p in soup.find_all("p"):
                pt = p.get_text()
                if len(buf) == 0 or pt not in buf[-1]:
                    buf.append(pt)
            return "\n".join(buf)
        except:
            return ""

    def _split_doc(
        self,
        text: str,
        max_words_per_paragrpah=384,
        short_paragraph_threshold=96,
        preserve_threshold=8,
    ) -> List[str]:
        """Use spacy to split a document to paragraphs."""
        paras = text.splitlines()
        splitted = []
        sent_to_be_concat = ""
        accumulate_length = 0
        for p in paras:
            p = p.strip()
            if len(p) < 1:
                continue  # empty lines
            sents = self.extractor.extract(p)
            for sent in sents:
                if accumulate_length + len(sent) <= max_words_per_paragrpah:
                    sent_to_be_concat += sent.text_with_ws
                    accumulate_length += len(sent)
                else:
                    splitted.append(sent_to_be_concat)
                    sent_to_be_concat = sent.text_with_ws
                    accumulate_length = len(sent)
            if accumulate_length <= short_paragraph_threshold:
                sent_to_be_concat += " "
            else:
                splitted.append(sent_to_be_concat)
                sent_to_be_concat = ""
                accumulate_length = 0
        if accumulate_length >= preserve_threshold:
            splitted.append(sent_to_be_concat)
        return splitted

    def _process_retrieved_docs(
        self,
        docs: List[dict],
        query: str,
        best_k=8,
        max_words_per_paragraph=384,
        skip_repeated_corpus=True,
    ) -> List[Dict[str, Union[str, None]]]:  # {"content": <text>, "url": <url>}
        if len(docs) == 0:
            return None
        if len(docs) == 1:
            return docs
        else:
            links_dict = {}
            corpus, links = [], []  # List of documents
            # retrieve through the links
            for relevance in docs:
                url = relevance.get("source", '')
                if not url:
                    continue
                if "youtube" in url:
                    continue  # skip youtube due to slow fetching
                if url in links_dict.keys():
                    if skip_repeated_corpus:
                        continue
                    online_text = links_dict[url]
                else:
                    online_text = self._get_url_text(url)
                    links_dict[url] = online_text
                splitted_text = self._split_doc(
                    online_text, max_words_per_paragraph
                )
                corpus.extend(splitted_text)
                links.extend([url] * len(splitted_text))

            meta_doc_dict = dict(zip(corpus, links))
            tokenized_corpus = [doc.split(" ") for doc in corpus]
            if not tokenized_corpus:
                return []
            bm25 = rank_bm25.BM25Okapi(tokenized_corpus)
            best_docs = bm25.get_top_n(query.split(), corpus, n=best_k)
            return [
                {"content": k, "source": meta_doc_dict[k]}
                for k in best_docs
            ]

    def retrieve(
        self,
        text: str,
        top_k=3,
        max_words_per_paragraph=384
    ) -> List[Dict[str, Union[str, None]]]:
        """
        Search reference documents on the Internet based on LLM generated query.
        Parameters
        ----------
        text : str
            Text to be checked.
        top_k : int
            Number of reference documents to be retrieved.
        max_words_per_paragraph : int
            Maximum number of words in each reference document.
        Returns
        -------
        List[str]
            List of reference documents
        """

        # Step 1. Generate queries for searching using LLM.
        queries = text
        # Step 2. Search google with the queries.
        relevant_info_dicts, best_docs_all = [], []
        for q in queries:
            searched_results = self._query_google(q)
            parsed_results, with_answerbox = self._parse_results(
                searched_results
            )
            if with_answerbox:
                answerbox_answer, parsed_results = (
                    parsed_results[0],
                    parsed_results[1:],
                )
            relevant_info_dicts.extend(parsed_results)
            best_docs = self._process_retrieved_docs(
                relevant_info_dicts,
                q,
                best_k=math.ceil((top_k - with_answerbox) / len(queries)),
                max_words_per_paragraph=max_words_per_paragraph,
                skip_repeated_corpus=True,
            )
            if with_answerbox:
                best_docs.insert(0, answerbox_answer)
            best_docs_all.extend(best_docs)
        refs = [
            doc["content"] for doc in best_docs_all
        ]
        return refs



from .extractor_base import ExtractorBase
from ..utils import get_claude2_response


CLAUDE2_TRIPLET_EXTRACTION_PROMPT_Q = \
"""Given a question and a candidate answer to the question, please extract a KG from the candidate answer condition on the question and represent the KG with triples formatted with ("subject", "predicate", "object"), each triplet in a line.
Please note that this is an EXTRACTION task, so DO NOT care about whether the content of the candidate answer is factual or not, just extract the triplets from it.

Here are some in-context examples:

### Question:
Given these paragraphs about the Tesla bot, what is its alias?

### Candidate Answer:
Optimus (or Tesla Bot) is a robotic humanoid under development by Tesla, Inc. It was announced at the company's Artificial Intelligence (AI) Day event on August 19, 2021.

### KG:
("Optimus", "is", "robotic humanoid")
("Optimus", "under development by", "Tesla, Inc.")
("Optimus", "also known as", "Tesla Bot")
("Tesla, Inc.", "announced", "Optimus")
("Announcement of Optimus", "occurred at", "Artificial Intelligence (AI) Day event")
("Artificial Intelligence (AI) Day event", "held on", "August 19, 2021")
("Artificial Intelligence (AI) Day event", "organized by", "Tesla, Inc.")

### Question:
here is some text about Andre Weiss, how many years was Andre at University of Dijon in Paris?

### Candidate Answer:
11 years

### KG:
("Andre Weiss at University of Dijon in Paris", "duration", "11 years")


Now generate the KG for the following candidate answer based on the provided question:

### Question:
{q}

### Candidate Answer:
{a}

### KG:
"""

CLAUDE2_TRIPLET_EXTRACTION_PROMPT = \
"""Given an input text, please extract a KG from the text and represent the KG with triples formatted with ("subject", "predicate", "object"), each triplet in a line. Please note that this is an EXTRACTION task, so DO NOT care about whether the content of the candidate answer is factual or not, just extract the triplets from it.

Here are some in-context examples:

### Input:
Optimus (or Tesla Bot) is a robotic humanoid under development by Tesla, Inc. It was announced at the company's Artificial Intelligence (AI) Day event on August 19, 2021.

### KG:
("Optimus", "is", "robotic humanoid")
("Optimus", "under development by", "Tesla, Inc.")
("Optimus", "also known as", "Tesla Bot")
("Tesla, Inc.", "announced", "Optimus")
("Announcement of Optimus", "occurred at", "Artificial Intelligence (AI) Day event")
("Artificial Intelligence (AI) Day event", "held on", "August 19, 2021")
("Artificial Intelligence (AI) Day event", "organized by", "Tesla, Inc.")

### Input:
The song "Here Comes the Boom" was originally released by American rock band Nelly in 2002 for the soundtrack of the film "The Longest Yard."

### KG:
("The song 'Here Comes the Boom'", "originally released by", "American rock band Nelly")
("The song 'Here Comes the Boom'", "released in", "2002")
("The song 'Here Comes the Boom'", "featured in", "soundtrack of the film 'The Longest Yard'")
("American rock band Nelly", "released", "The song 'Here Comes the Boom'")
("The Longest Yard", "had soundtrack featuring", "The song 'Here Comes the Boom'")


Now generate the KG for the provided input text:

### Input:
{input_text}

### KG:
"""


class Claude2Extractor(ExtractorBase):
    def __init__(
        self,
        claim_format:str='triplet'
    ) -> None:
        super().__init__(claim_format=claim_format)
        
        if self.claim_format == 'triplet':
            self.prompt_temp_wq = CLAUDE2_TRIPLET_EXTRACTION_PROMPT_Q
            self.prompt_temp = CLAUDE2_TRIPLET_EXTRACTION_PROMPT

    def extract_claim_triplets(self, response, question=None, max_new_tokens=500):
        if question is None:
            prompt = self.prompt_temp.format(
                input_text=response
            )
        else:
            prompt = self.prompt_temp_wq.format(
                q=question,
                a=response
            )

        claude2_response = get_claude2_response(
            prompt=prompt,
            temperature=0,
            max_new_tokens=max_new_tokens
        )

        if claude2_response and len(claude2_response):
            kg_str = None
            if '###' in claude2_response:
                kg_str = claude2_response[:claude2_response.index('###')]
            else:
                kg_str = claude2_response
            triplets = self._parse_claim_triplets(kg_str)
            return triplets
        return []



# from .claude2_checker import Claude2Checker
from .gpt4_checker import GPT4Checker
# from .nli_checker import NLIChecker
# from .alignscore.alignscore_checker import AlignScoreChecker
# from .repc.repc_checker import RepCChecker




import logging
import os
from openai import OpenAI

logger = logging.getLogger(__name__)


class OpenAIHandler:
    def __init__(self):
        self.client = OpenAI(api_key=os.getenv('OPENAI_API_KEY'))

    def ask_llm(self, prompt, n=1, temperature=0, max_new_tokens=int(os.getenv("OPENAI_MAX_TOKENS", 400))):
        response = self.client.completions.create(
            model=os.getenv("OPENAI_MODEL"),
            prompt=prompt,
            max_tokens=max_new_tokens,
            n=n,
            temperature=temperature
        )
        results = [r.text.strip() for r in response.choices]
        logger.info(f'Prompt responses: {results}')
        return results



from .google_retriever import GoogleRetriever


class RetrieverHandler:
    def __init__(self, sentence_extractor):
        self.retriever = GoogleRetriever(sentence_extractor)

    def retrieve(self, *args, **kwargs):
        return self.retriever.retrieve(*args, **kwargs)



import spacy

from .gpt4_extractor import GPT4Extractor


class TripletsExtractorHandler:
    def __init__(self, llm_handler):
        self.extractor = GPT4Extractor(llm_handler)

    def extract(self, response, question=None, max_new_tokens=200):
        return self.extractor.extract_claim_triplets(response, question, max_new_tokens)


class SentenceExtractorHandler:
    def __init__(self):
        self.nlp = spacy.load("en_core_web_sm")

    def extract(self, text):
        return [sent for sent in self.nlp(text).sents]



import logging

PROMPT = """Please generate a question on the given text so that when searching on Google with the question, it's possible to get some relevant information on the topics addressed in the text. Note, you just need to give the final question without quotes in one line, and additional illustration should not be included.

For example:
Input text: The Lord of the Rings trilogy consists of The Fellowship of the Ring, The Two Towers, and The Return of the King.
Output: What are the three books in The Lord of the Rings trilogy?

Input text: %s
Output: """


logger = logging.getLogger(__name__)


class QuestionGenerator:
    def __init__(self, llm_handler):
        self.llm_handler = llm_handler

    def generate(self, paragraph):
        prompt = PROMPT % paragraph
        response = self.llm_handler.ask_llm(prompt)[0]
        logger.info(f'Question generated: {response}')
        return response



from .base import Detector


class RefChecker(Detector):
    def __init__(self):
        super().__init__()

    def score(self, question=None, answer=None, samples=None):
        if not question:
            question = self.generate_question(answer)

        if not answer:
            answer = self.ask_llm(question.strip())[0]

        triplets = self.extract_triplets(answer, question, max_new_tokens=200)
        reference = self.retrieve([question])

        results = [
            self.check(t, reference, question=question)
            for t in triplets
        ]
        agg_results = self.soft_agg(results)
        for k, v in agg_results.items():
            agg_results[k] = float("{:.2f}".format(v))
        return agg_results, answer, None

    def soft_agg(self, results):
        """Aggregate results by taking the ratio of each category."""
        if not results:
            return {
                "Entailment": 0.0,
                "Neutral": 0.0,
                "Contradiction": 0.0,
                "Abstain": 1.0,
            }
        total = len(results)
        agg = {
            "Entailment": 0.0,
            "Neutral": 0.0,
            "Contradiction": 0.0,
            "Abstain": 0.0,
        }
        for result in results:
            agg[result] += 1.0
        for key in agg:
            agg[key] /= total
        print(results)
        return agg



import logging
import os

from .base import Detector


logger = logging.getLogger(__name__)


class ChainPoll(Detector):

    def __init__(self):
        super().__init__()
        with open(f'{os.path.dirname(os.path.realpath(__file__))}/../prompts/chainpoll.txt', 'r') as pf:
            self.prompt = pf.read()

    def check_hallucinations(self, completion, question):
        text = self.prompt.format(completion=completion, question=question)
        responses = self.ask_llm(text, n=int(os.getenv("CHAINPOLL_SAMPLING_NUMBER")), temperature=0.2)
        logger.info(f'Hallucination check response: {responses}')
        return [response.lower().startswith("yes") for response in responses], responses

    def score(self, question, answer=None, samples=None):
        if not answer:
            answer = self.ask_llm(question.strip())[0]
        hallucinations, responses = self.check_hallucinations(answer.strip(), question.strip())

        score = hallucinations.count(True) / len(hallucinations)
        return score, answer, responses



import logging
import os
from transformers import AutoModelForCausalLM, AutoTokenizer

logger = logging.getLogger(__name__)


class MistralHandler:
    def __init__(self):
        self.model = AutoModelForCausalLM.from_pretrained("mistralai/Mistral-7B-v0.1", device_map="auto")
        self.tokenizer = AutoTokenizer.from_pretrained("mistralai/Mistral-7B-v0.1")

    def ask_llm(self, prompt, n=1, temperature=0, max_new_tokens=int(os.getenv("OPENAI_MAX_TOKENS", 400))):
        model_inputs = self.tokenizer([prompt] * n, return_tensors="pt")
        generated_ids = self.model.generate(**model_inputs, max_new_tokens=max_new_tokens, do_sample=True)
        results = [r for r in self.tokenizer.batch_decode(generated_ids)]
        logger.info(f'Prompt responses: {results}')
        return results



from .chainpoll import ChainPoll
from .selfcheckgpt import SelfCheckGPTBertScore, SelfCheckGPTNGram, SelfCheckGPTMQAG, SelfCheckGPTPrompt
from .refchecker import RefChecker
from .geval import GEval
from .custom import CustomDetector





from .covid_qa import CovidQAParser
from .databricks_dolly import DollyParser
from .drop import DropParser
from .parser import Parser

BENCHMARKS = [
    CovidQAParser,
    DropParser,
    DollyParser
]


def benchmarks_for_UI():
    return [{'id': benchmark._id, 'display_name': benchmark.display_name} for benchmark in BENCHMARKS]


def get_benchmark(_id):
    for benchmark in BENCHMARKS:
        if benchmark._id == _id:
            return benchmark



import json


class Parser:
    display_name = None
    _id = None

    def parse_input(self, form):
        data = json.loads(form['jsonData'])
        selected_indexes = json.loads(form['selectedIndexes'])
        return [res for index, res in enumerate(data['data']) if index in selected_indexes]



import logging
import os
import spacy
import numpy as np
from selfcheckgpt.modeling_selfcheck import SelfCheckMQAG

from .base import Detector


logger = logging.getLogger(__name__)


class SelfCheckGPTBertScore(Detector):

    def __init__(self):
        super().__init__()
        self.sample_number = int(os.getenv("BERT_SCORE_SAMPLING_NUMBER", 3))

    def score(self, question, answer=None, samples=None):
        if not answer:
            answer = self.ask_llm(question)[0]

        sentences = self.extract_sentences(answer)
        sentences = [s.text for s in sentences]
        if not samples:
            samples = self.ask_llm(question, n=self.sample_number, temperature=0.4)

        scores = self.similarity_bertscore(
            sentences=sentences,  # list of sentences
            sampled_passages=samples,  # list of sampled passages
        )
        if hasattr(scores, '__iter__'):
            scores = float("{:.2f}".format(sum(scores)/len(scores)))
        else:
            scores = 'FAILED'
        return scores, answer, samples


class SelfCheckGPTNGram(Detector):

    def __init__(self):
        super().__init__()
        self.sample_number = int(os.getenv("NGRAM_SAMPLING_NUMBER", 3))

    def score(self, question, answer=None, samples=None):
        if not answer:
            answer = self.ask_llm(question)[0]

        sentences = self.extract_sentences(answer)
        sentences = [s.text for s in sentences]
        if not samples:
            samples = self.ask_llm(question, n=self.sample_number, temperature=0.4)

        scores = self.similarity_ngram(
            sentences=sentences,
            passage=answer,
            sampled_passages=samples
        )

        scores = float("{:.2f}".format(scores['doc_level']['avg_neg_logprob']))
        return scores, answer, samples


class SelfCheckGPTPrompt(Detector):

    def __init__(self):
        super().__init__()
        self.sample_number = int(os.getenv("PROMPT_SAMPLING_NUMBER", 3))
        self.prompt_template = "Context: {context}\n\nSentence: {sentence}\n\nIs the sentence supported by the context above? Answer Yes or No.\n\nAnswer: "
        self.text_mapping = {'yes': 0.0, 'no': 1.0, 'n/a': 0.5}

    def score(self, question, answer=None, samples=None):

        if not answer:
            answer = self.ask_llm(question)[0]

        if not samples:
            samples = self.ask_llm(question, n=self.sample_number, temperature=0.4)

        sentences = self.extract_sentences(answer)
        sentences = [s.text for s in sentences]
        scores = np.zeros((len(sentences), len(samples)))
        for sent_i, sentence in enumerate(sentences):
            sentence = sentences[sent_i]
            for sample_i, sample in enumerate(samples):
                sample = sample.strip()
                prompt = self.prompt_template.format(context=sample, sentence=sentence)
                generate_text = self.ask_llm(prompt)[0]
                generate_text = generate_text.replace(prompt, "")
                score_ = self.text_postprocessing(generate_text)
                scores[sent_i, sample_i] = score_
        scores_per_sentence = scores.mean(axis=-1)
        return sum(scores_per_sentence)/len(scores_per_sentence), answer, samples

    def set_prompt_template(self, prompt_template: str):
        self.prompt_template = prompt_template

    def text_postprocessing(
        self,
        text,
    ):
        """
        To map from generated text to score
        Yes -> 0.0
        No  -> 1.0
        everything else -> 0.5
        """
        # tested on Llama-2-chat (7B, 13B) --- this code has 100% coverage on wikibio gpt3 generated data
        # however it may not work with other datasets, or LLMs
        text = text.lower().strip()
        if text[:3] == 'yes':
            text = 'yes'
        elif text[:2] == 'no':
            text = 'no'
        else:
            if text not in self.not_defined_text:
                print(f"warning: {text} not defined")
                self.not_defined_text.add(text)
            text = 'n/a'
        return self.text_mapping[text]


class SelfCheckGPTMQAG(Detector):

    def __init__(self):
        import torch
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.scorer = SelfCheckMQAG(device=device)
        self.sample_number = 3

    def score(self, question, answer=None, samples=None):
        samples = []
        if not answer:
            answer = self.ask_llm(question)[0]

        nlp = spacy.load("en_core_web_sm")
        sentences = [sent.text.strip() for sent in nlp(answer).sents]

        for _ in range(self.sample_number):
            samples.append(self.ask_llm(question)[0])

        scores = self.scorer.predict(
            sentences=sentences,  # list of sentences
            passage=answer,  # passage (before sentence-split)
            sampled_passages=samples,  # list of sampled passages
            num_questions_per_sent=5,  # number of questions to be drawn
            scoring_method='bayes_with_alpha',  # options = 'counting', 'bayes', 'bayes_with_alpha'
            beta1=0.8, beta2=0.8,
        )
        scores = float("{:.2f}".format(sum(scores)/len(scores)))
        return scores, answer, samples



import json
import os

from .parser import Parser


class CovidQAParser(Parser):
    display_name = 'Covid-QA'
    _id = 'covid-qa'

    def __init__(self, file='../../datasets/Covid-QA.json'):
        self.file = file
        if self.file.startswith('..'):
            self.file = f'{os.path.dirname(os.path.realpath(__file__))}/{file}'
        with open(self.file, 'r') as parsefile:
            self.data = json.loads(parsefile.read())

    def display(self):
        results = []
        for data in self.data['data']:
            for paragraph in data['paragraphs']:
                for qas in paragraph['qas']:
                    results.append({'question': qas['question'], 'id': qas['id'], 'answer': qas['answers'][0]['text']})
        return {
            'data': results,
            'columns': ['id', 'question', 'answer']
        }



from .extractor_base import ExtractorBase

import torch
from vllm import LLM, SamplingParams


one_digit_tensor = torch.ones((1, 1), dtype=torch.long)

MISTRAL_KG_EXTRACTION_PROMPT_Q = """Given a question and a candidate answer to the question, please extract a KG from the candidate answer condition on the question and represent the KG with triples formatted with ("head", "relation", "tail"), each triplet in a line.
Please note that this is an EXTRACTION task, so DO NOT care about whether the content of the candidate answer is factual or not, just extract the triplets from it.
Now generate the KG for the following candidate answer based on the provided question:

### Question:
{q}

### Candidate Answer:
{a}

### KG:
"""


class MistralExtractor(ExtractorBase):
    def __init__(
        self,
        claim_format: str = "triplet",
        tensor_parallel_size=-1,
    ) -> None:
        super().__init__(claim_format=claim_format)
        self.llm = LLM(
            "dongyru/Mistral-7B-Claim-Extractor",
            load_format="safetensors",
            tensor_parallel_size=torch.cuda.device_count()
            if tensor_parallel_size < 0
            else tensor_parallel_size,
            trust_remote_code=True,
        )
        self.tokenizer = self.llm.get_tokenizer()
        if self.claim_format == "triplet":
            self.prompt_temp_wq = MISTRAL_KG_EXTRACTION_PROMPT_Q
            self.prompt_temp = None


    # TODO move this to utils for general LLMs based on vllm interface
    def _get_response_from_mistral(
        self, raw_prompt: str, max_new_tokens: int = 2048
    ):
        llm = self.llm
        tokenizer = self.tokenizer
        prompt_token_ids = tokenizer.encode(f"[INST] {raw_prompt} [/INST]")
        outputs = llm.generate(
            prompt_token_ids=[prompt_token_ids],
            sampling_params=SamplingParams(
                temperature=0., max_tokens=max_new_tokens
            ),
            use_tqdm=False,
        )
        llm_output = outputs[0].outputs[0].text
        return llm_output

    def extract_claim_triplets(
        self, response, question=None, max_new_tokens=2048
    ):
        if question is None:
            raise NotImplementedError(
                "Question is required for MistralExtractor"
            )
        else:
            prompt = self.prompt_temp_wq.format(q=question, a=response)
        mistral_response = self._get_response_from_mistral(
            prompt, max_new_tokens=max_new_tokens
        )
        if mistral_response and len(mistral_response):
            kg_str = None
            if "###" in mistral_response:
                kg_str = mistral_response[: mistral_response.index("###")]
            else:
                kg_str = mistral_response
            triplets = self._parse_claim_triplets(kg_str)
            return triplets
        return []


if __name__ == "__main__":
    import json

    extractor = MistralExtractor()
    example_for_test = {
        "question": "full time student how many hours",
        "response": 'Based on the provided passages, the number of hours required to be considered a full-time student can vary depending on the context. However, some common requirements mentioned are:\n\n- Passage 0: Full-time status is usually considered as a schedule of 12 or more semester or quarter hours, but there is no data on how many credits full-time students are actually taking.\n\n- Passage 1: For undergraduate students in the summer, 12 hours is considered full-time.\n\n- Passage 2: For fall and spring semesters, a full-time college student completes at least 12 semester hours. Some schools may require 15 semester hours. In the summer, completing 6 semester hours is considered full-time.\n\n- Passage 3: For graduate students, a normal full-time load is nine graduate-level semester hours. But for those holding graduate assistantships, full-time status is six semester hours.\n\n- Passage 4: According to the University Bulletin, graduate students taking 9 or more credit hours per semester (6 credits in the summer) are considered full-time.\n\n- Passage 5: The normal load for full-time students is 3 courses (9 credits).\n\n- Passage 8: For graduate students, 9 hours is considered full-time.\n\nBased on the information provided, the answer to the question "full time student how many hours?" is not explicitly stated or consistent across all passages. It can range from 12 hours for undergraduate students in the summer to 9 or more credit hours for graduate students. Some schools may require 15 semester hours for full-time status.',
    }

    print(
        json.dumps(
            extractor.extract_claim_triplets(
                response=example_for_test["response"],
                question=example_for_test["question"]
            ),
            indent=4,
        )
    )



from datasets import load_dataset
from .parser import Parser


class DollyParser(Parser):
    display_name = 'Databricks Dolly'
    _id = 'databricks-dolly'

    def __init__(self):
        self.dataset = load_dataset('databricks/databricks-dolly-15k')
        self.dataset = self.dataset['train']

    def display(self):
        results = []

        for element in self.dataset:
            results.append(
                {
                    'question': element['instruction'],
                    'context': element['context'],
                    'answer': element['response'],
                    'category': element['category']
                }
            )
        return {
            'data': results,
            'columns': ['question', 'context', 'answer', 'category']
        }



import re


class ExtractorBase:
    def __init__(
        self,
        claim_format:str='triplet'
    ) -> None:
        self.claim_format = claim_format

    def extract(self, response, question=None, max_new_tokens=500):
        claims = None
        if self.claim_format == 'triplet':
            claims = self.extract_claim_triplets(
                response=response,
                question=question,
            )
        return claims

    def extract_claim_triplets(
        self,
        response,
        question=None, 
        max_new_tokens=500
    ):
        raise NotImplementedError

    def _parse_claim_triplets(self, text):
        ret = []
        patterns = [
            r'\(".*", ".*", ".*"\)',
            r'\(".*", ".*", \'.*\'\)',
            r'\(".*", \'.*\', ".*"\)',
            r'\(\'.*\', ".*", ".*"\)',
            r'\(".*", \'.*\', \'.*\'\)',
            r'\(\'.*\', ".*", \'.*\'\)',
            r'\(\'.*\', \'.*\', ".*"\)',
            r'\(\'.*\', \'.*\', \'.*\'\)'
        ]
        for p in patterns:
            triplets = self._parse_triplets(p, text, triple_length=3)
            if triplets:
                ret += triplets

        # deduplication
        final_triple_set = []
        for t in ret:
            if tuple(t) not in final_triple_set:
                final_triple_set.append(tuple(t))
        
        return [list(t) for t in final_triple_set]

    def _parse_triplets(self, pattern, text, triple_length=3):
        triplets = []
        matches = re.findall(pattern, text)
        for m in matches:
            try:
                t = eval(m)
            except:
                t = m.split(', ')
                if t[0].startswith('('):
                    t[0] = t[0][1:]
                if t[-1].endswith(')'):
                    t[-1] = t[-1][:-1]
            if len(t) != triple_length:
                continue
            if any([not isinstance(e, str) for e in t]):
                continue
            if any([len(e) == 0 for e in t]):
                continue
            triplets.append(list(t))
        return triplets



from setuptools import find_packages, setup


def parse_requirements(filename):
    with open(filename, 'r') as f:
        return [line.strip() for line in f if line.strip() and not line.startswith("#")]


requirements = parse_requirements('requirements.txt')


setup(
    name='halludetector',
    version='0.0.2',
    author='Mihai Onofrei',
    author_email='monofrei@cisco.com',
    url='https://github.com/Mihai-Onofrei/Hallucination-detector',  # Replace with your project's URL
    license='MIT',  # Choose an appropriate license
    description="Hallucination detection package",
    long_description=open('README.rst').read(),
    include_package_data=True,
    packages=find_packages(),
    package_data={
        '': ['*.json'],  # Include all JSON files in the package
        'logos': ['*.png'],
        'js': ['*.js'],
        'txt': ['*.txt']
    },
    install_requires=requirements
)



from datasets import load_dataset

from .parser import Parser


class DropParser(Parser):
    display_name = 'Drop'
    _id = 'drop'

    def __init__(self):
        self.dataset = load_dataset('EleutherAI/drop')
        self.dataset = self.dataset['validation']

    def display(self):
        results = []

        for element in self.dataset:
            answer = ''
            answer_element = element['answer']
            if answer_element['number']:
                answer = answer_element['number']
            elif answer_element['date']['day']:
                answer = f'{answer_element["date"]["year"]}-{answer_element["date"]["month"]}-{answer_element["date"]["day"]}'
            elif answer_element['spans']:
                answer = ' ,'.join(answer_element['spans'])

            samples = []
            for ans in element['validated_answers']['number']:
                if ans:
                    samples.append(ans)
            for ans in element['validated_answers']['date']:
                if ans['day']:
                    samples.append(f'{ans["year"]}-{ans["month"]}-{ans["day"]}')
            for ans in element['validated_answers']['spans']:
                if ans:
                    samples.append(' ,'.join(ans))

            results.append(
                {
                    'question': element['question'],
                    'context': element['passage'],
                    'answer': answer,
                    'samples': samples,
                    'query_id': element['query_id']
                }
            )
        return {
            'data': results,
            'columns': ['question', 'context', 'answer', 'samples', 'query_id']
        }



import logging
import os

import re

from .base import Detector


logger = logging.getLogger(__name__)


class GEval(Detector):
    metrics = ['coherence', 'consistency', 'fluency', 'relevance']
    @staticmethod
    def parse_output(output):
        matched = re.search("^ ?([\d\.]+)", output)
        if matched:
            try:
                score = float(matched.group(1))
            except:
                score = 0
        else:
            score = 0
        return score

    @staticmethod
    def normalize_score(score):
        max_score = 5  # Maximum possible score
        normalized_score = score / max_score
        return normalized_score

    def create_prompt(self, answer, summary, metric=os.getenv("GEVAL_METRIC")):
        mapping = {
            "coherence": "coh_detailed.txt",
            "consistency": "con_detailed.txt",
            "fluency": "flu_detailed.txt",
            "relevance": "rel_detailed.txt"
        }
        with open(os.path.dirname(os.path.abspath(__file__)) + f'/../prompts/{mapping[metric]}') as readfile:
            prompt = readfile.read()
        cur_prompt = prompt.replace('{{Document}}', answer).replace('{{Summary}}', summary)
        return cur_prompt

    def score(self, question, answer=None, samples=None):
        scores = {}
        if not answer:
            answer = self.ask_llm(question)[0]
        answer = answer.strip()
        summary_prompt = f"Create a summary with 20 maximum words from {answer}"
        summary = self.ask_llm(summary_prompt)[0].strip()

        for metric in self.metrics:
            prompt = self.create_prompt(answer, summary, metric)
            answers = self.ask_llm(prompt, n=int(os.getenv("GEVAL_SAMPLING_NUMBER", 5)))

            all_scores = [self.parse_output(x.strip()) for x in answers]
            score = sum(all_scores) / len(all_scores)
            scores[metric.title()] = float("{:.2f}".format(self.normalize_score(score)))
        return scores, answer, []



import logging

logging.basicConfig()

# By default the root logger is set to WARNING and all loggers you define
# inherit that value. Here we set the root logger to NOTSET. This logging
# level is automatically inherited by all existing and new sub-loggers
# that do not set a less verbose level.
logging.root.setLevel(logging.INFO)

# The following line sets the root logger level as well.
# It's equivalent to both previous statements combined:
logging.basicConfig(level=logging.INFO)
logging.getLogger("httpx").setLevel(logging.WARNING)

from .scorer import calculate_score
from .config import init_config



import json
import os

from selfcheckgpt.modeling_selfcheck import SelfCheckBERTScore, SelfCheckNgram

from halludetector.llm.openai import OpenAIHandler
from halludetector.llm.mistral import MistralHandler
from halludetector.extractor.extractor import TripletsExtractorHandler, SentenceExtractorHandler
from halludetector.generators.question import QuestionGenerator
from halludetector.retrievers.retriever import RetrieverHandler
from halludetector.checker.checker import CheckerHandler

llm_handler = None
triplets_extractor = None
sentence_extractor = None
question_generator = None
retriever = None
checker = None
bertscorer = None
ngramscorer = None


def init_config(file, force=False):
    with open(file, 'r') as cfgfile:
        config = json.loads(cfgfile.read())
        for c in config:
            key = c['name']
            value = c['value']
            if force or not os.getenv(key):
                os.environ[key] = str(value)
    init_building_blocks()


def init_building_blocks(force=False):
    global llm_handler, triplets_extractor, sentence_extractor, question_generator, \
        retriever, checker, bertscorer, ngramscorer

    if llm_handler is None or force:
        llm_handler = OpenAIHandler()
        triplets_extractor = TripletsExtractorHandler(llm_handler)
        sentence_extractor = SentenceExtractorHandler()
        question_generator = QuestionGenerator(llm_handler)
        retriever = RetrieverHandler(sentence_extractor)
        checker = CheckerHandler(sentence_extractor, llm_handler)
        bertscorer = SelfCheckBERTScore(rescale_with_baseline=True)
        ngramscorer = SelfCheckNgram(n=int(os.getenv("NGRAM_N", 1)))



from typing import List

from .checker_base import CheckerBase


GPT4_CHECKING_PROMPT_Q = \
"""I have a claim that made by a language model to a question, please help me for checking whether the claim can be entailed according to the provided reference which is related to the question. 
The reference is a list of passages, and the claim is represented as a triplet formatted with ("subject", "predicate", "object").

If the claim is supported by ANY passage in the reference, answer 'Entailment'. 
If NO passage in the reference entail the claim, and the claim is contradicted with some passage in the reference, answer 'Contradiction'.
If NO passage entail or contradict with claim, or DOES NOT contain information to verify the claim, answer 'Neutral'. 

Please DO NOT use your own knowledge for the judgement, just compare the reference and the claim to get the answer.

### Question:
{question}

### Reference:
{reference}

### Claim:
{claim}

Your answer should always be only a single word in ['Entailment', 'Neutral', 'Contradiction']. DO NOT add explanations or you own reasoning to the output.
"""

GPT4_CHECKING_PROMPT = \
"""I have a claim that made by a language model, please help me for checking whether the claim can be entailed according to the provided reference. 
The reference is a list of passages, and the claim is represented as a triplet formatted with ("subject", "predicate", "object").

If the claim is supported by ANY passage in the reference, answer 'Entailment'. 
If NO passage in the reference entail the claim, and the claim is contradicted with some passage in the reference, answer 'Contradiction'.
If NO passage entail or contradict with claim, or DOES NOT contain information to verify the claim, answer 'Neutral'. 

Please DO NOT use your own knowledge for the judgement, just compare the reference and the claim to get the answer.

### Reference:
{reference}

### Claim:
{claim}

Your answer should always be only a single word in ['Entailment', 'Neutral', 'Contradiction']. DO NOT add explanations or you own reasoning to the output.
"""


class GPT4Checker(CheckerBase):
    def __init__(self, sentence_extractor, llm_handler) -> None:
        super().__init__(sentence_extractor)
        self.prompt_temp = GPT4_CHECKING_PROMPT
        self.prompt_temp_wq = GPT4_CHECKING_PROMPT_Q
        self.llm_handler = llm_handler

    def _check(
        self, 
        claims: List, 
        references: List,
        response: str,
        question: str = None,
    ):
        ret_labels = []
        for claim, reference in zip(claims, references):
            if isinstance(claim, list):
                assert len(claim) == 3
                claim = f"({claim[0]}, {claim[1]}, {claim[2]})"
            if question is None:
                prompt = self.prompt_temp.format(
                    reference=reference,
                    claim=claim
                )
            else:
                prompt = self.prompt_temp_wq.format(
                    question=question,
                    reference=reference,
                    claim=claim
                )
            openai_response = self.llm_handler.ask_llm(
                prompt=prompt,
            )[0]
            if openai_response and len(openai_response):
                label = None
                if self.label_contradiction.lower() in openai_response.lower():
                    label = self.label_contradiction
                elif self.label_entailment.lower() in openai_response.lower():
                    label = self.label_entailment
                else:
                    label = self.label_neutral
                ret_labels.append(label)
            else:
                raise 'OpenAI API returns None or empty string'
        return ret_labels



from .base import Detector


class CustomDetector(Detector):
    def score(self, question, answer=None, samples=None):
        # do your logic here using the building blocks from the Detector class
        # or use your own
        score = 0.0
        return score, answer, samples
